import numpy as np
import h5py

from collections import namedtuple
from extra_data.exceptions import PropertyNameError
from extra_redu.base import (
    AgipdLitFrameFinderBase, ReferenceDelay, AGIPDGEN, EVENT_PULSE,
    AGIPD_MAX_CELL, DESTINATION, SA1, SA2, SA3
)


# Exceptions
class LitFrameFinderError(RuntimeError):
    pass


class SourceNotFound(LitFrameFinderError):
    pass


class BunchPatternNotFound(SourceNotFound):
    pass


class DetectorNotFound(SourceNotFound):
    pass


class TriggerNotFound(SourceNotFound):
    pass


class NoReferenceDelay(LitFrameFinderError):
    pass


AttributeValue = namedtuple('AttributeValue', ['value'])
AVal = AttributeValue


def mangle_device_id(device_id):
    return ''.join(c if c.isalnum() else '_' for c in device_id)


class ExtraDataProxy:
    def __init__(self, dc, dev_id, node=None, data_selector_id=None):

        if dev_id in dc.control_sources:
            self._mangled_dev_id = None
            self._src = dc._get_source_data(dev_id)
            self._src_path = "RUN/" + dev_id
            self._check_path = self._check_device_path
            self._read_run_value = self._get_run_value_from_device
        elif data_selector_id in dc.control_sources:
            self._mangled_dev_id = mangle_device_id(dev_id)
            self._src = dc._get_source_data(data_selector_id)
            self._src_path = '/'.join([
                "RUN", data_selector_id, self._mangled_dev_id])
            self._check_path = self._check_data_selector_path
            self._read_run_value = self._get_run_value_from_data_selector
        else:
            raise SourceNotFound(f"{dev_id} is not found")

        if not self._check_path(self._src, self._src_path, node):
            path = self._src_path
            if node is not None:
                path += '/' + node.replace('.', '/')
            raise SourceNotFound(f"{path} is not found")

        self._dc = dc
        self._dev_id = dev_id
        self._data_selector_id = data_selector_id
        self._node = node
        self._cache_run_value = dict()
        if self._mangled_dev_id and self._node is None:
            self._cache_run_value['deviceId'] = AVal(dev_id)

    def _check_device_path(self, src, src_path, key):
        path = src_path
        if key is not None:
            path += '/' + key.replace('.', '/')
        return path in src.files[0].file

    def _check_data_selector_path(self, src, path, key):
        if path not in src.files[0].file:
            return False
        if key is None:
            return True
        grp = src.files[0].file[path]
        key_name = key.replace('.', '_')
        return any(ds_name.startswith(key_name) for ds_name in grp)

    def get(self, name):
        try:
            return self._cache_run_value[name]
        except KeyError:
            v = self._read_run_value(name)
            self._cache_run_value[name] = v
            return v

    def get_value(self, name):
        return self.get(name).value

    def __getattr__(self, name):
        return self.get(name)

    def _get_run_value_from_device(self, name):
        key = name if self._node is None else self._node + '.' + name
        if not self._check_device_path(self._src, self._src_path, key):
            raise AttributeError(
                f"'{self._dev_id}' does not have property '{name}'")

        path = self._src_path + '/' + key.replace('.', '/')
        grp = self._src.files[0].file[path]
        if isinstance(grp, h5py.Group) and ('value' not in grp):
            return ExtraDataProxy(
                self._dc, self._dev_id, key, self._data_selector_id)
        try:
            # replace to self._src.run_value(key)
            return AVal(self._dc.get_run_value(self._dev_id, key))
        except Exception:
            raise AttributeError(
                f"'{self._dev_id}' does not have property '{name}'")

    def _get_run_value_from_data_selector(self, name):
        key = name if self._node is None else self._node + '.' + name
        if not self._check_data_selector_path(self._src, self._src_path, key):
            raise AttributeError(
                f"'{self._dev_id}' does not have property '{name}'")

        key_path = key.replace('.', '_')
        path = self._src_path + '/' + key_path
        if path not in self._src.files[0].file:
            return ExtraDataProxy(
                self._dc, self._dev_id, key, self._data_selector_id)
        try:
            mangled_key = self._mangled_dev_id + '/' + key_path
            # replace to self._src.run_value(key)
            return AVal(self._dc.get_run_value(
                self._data_selector_id, mangled_key))
        except Exception:
            raise AttributeError(
                f"'{self._dev_id}' does not have property '{name}'")


Trigger = namedtuple('Trigger', ['deviceId', 'name'])
Motor = namedtuple('Motor', ['deviceId', 'positionProperty',
                             'unitConversionCoef'])
ReferenceListNode = namedtuple('ReferenceListNode', ['dynamic', 'static'])


class AgipdLitFrameFinderOffline(AgipdLitFrameFinderBase):

    data_selector_id = None

    alignMethod = AVal('by first pulse')
    referenceFrame = AVal(1)

    missedIntensityTD = AVal(10000)
    missedIntensitySigmaTD = AVal(0)

    useDistanceDelay = AVal(False)
    referenceDelay = ReferenceListNode(**dict([
        (event_name, ReferenceDelay(-1, 0, 1, 4.5, 0.0))
        for event_name in ReferenceListNode._fields
    ]))

    def __init__(self, dev_id, dc, ref_delays={}, align_method=None, **kwargs):
        self.dev_id = dev_id
        self.dc = dc
        self._reference = ref_delays

        for a, v in kwargs.items():
            setattr(self, a, AVal(v))

        if dev_id in dc.control_sources:
            try:
                self._configure(dc, dev_id)
            except (PropertyNameError, AttributeError):
                self._configure_v03(dc, dev_id)

        try:
            self._detector = ExtraDataProxy(
                dc, self.detector_control_id,
                data_selector_id=None  # classId missed in the data selector
            )
        except SourceNotFound:
            raise(DetectorNotFound(
                f"Source '{self.detector_control_id}' is not found."))

        if align_method is not None:
            self.alignMethod = AVal(align_method)

        # inst = self.instrumentName.value
        inst = self._detector.deviceId.value[:3]
        self.dst = DESTINATION[inst]

        self._get_event_pulse = {}
        for event_name, event_code in self.dst.event_code.items():
            ref = getattr(self.referenceDelay, event_name)
            if ref.delay == -1:
                continue
            self._reference[event_code] = ref
            self._get_event_pulse[event_code] = EVENT_PULSE[event_name]

        agipd_class = AGIPDGEN.get(self._detector.classId.value)
        if agipd_class is None:
            raise ValueError(
                f"Unknown class '{agipd_class}' of AgipdComposite device")
        self._agipd = agipd_class(self._detector, self)

        self.shutters_are_open = True
        self.actual_position = None

    def _configure_v03(self, dc, dev_id):
        litfrm = ExtraDataProxy(dc, dev_id)

        self.detector_control_id = litfrm.DetectorCtrl.value
        self.alignMethod = litfrm.alignMethod
        self.instrumentName = litfrm.instrumentName
        self.referenceFrame = litfrm.referenceFrame

        self.trigger = Trigger(
            deviceId=litfrm.Timer.value,
            name=litfrm.trigger,
        )
        self.motor = Motor(
            deviceId='',
            positionProperty='',
            unitConversionCoef=1,
        )

    def _configure(self, dc, dev_id):
        litfrm = ExtraDataProxy(dc, dev_id)

        self.detector_control_id = litfrm.detectorId.value
        self.alignMethod = litfrm.alignMethod
        self.instrumentName = litfrm.instrumentName
        self.referenceFrame = litfrm.referenceFrame

        self.trigger = Trigger(
            deviceId=litfrm.trigger.deviceId.value,
            name=litfrm.trigger.name,
        )
        self.motor = Motor(
            deviceId=litfrm.motor.deviceId.value,
            positionProperty=litfrm.motor.positionProperty,
            unitConversionCoef=litfrm.motor.unitConversionCoef,
        )
        reference_delays = {}
        for event_name in ReferenceListNode._fields:
            reference_delays[event_name] = ReferenceDelay(**dict([
                (key, litfrm.get(f"referenceDelay.{event_name}.{key}").value)
                for key in ReferenceDelay._fields
            ]))
        self.referenceDelay = ReferenceListNode(**reference_delays)
        try:
            self.useDistanceDelay = litfrm.useDistanceDelay
            self.detectorDistance = litfrm.detectorDistance
        except AttributeError:
            self.useDistanceDelay = AVal(True)
            self.detectorDistance = AVal(0.0)

    def _read_run_values(self, dc):
        if self.alignMethod.value == "by reference delay":
            self._read_trigger_and_motor_values(dc)

    def _read_trigger_and_motor_values(self, dc):
        try:
            _timer = ExtraDataProxy(dc, self.trigger.deviceId,
                                    data_selector_id=self.data_selector_id)
        except SourceNotFound:
            raise TriggerNotFound(
                f"Source {self.trigger.deviceId} is not found.")

        key = self.trigger.name.value
        self.trigger_event = _timer.get_value(key + '.event')
        self.trigger_delay = _timer.get_value(key + '.delay')

        self.ref = self._reference.get(self.trigger_event)
        if self.ref is None:
            raise(NoReferenceDelay("Reference delay for marco event "
                                   f"{self.trigger_event} is not set."))

        if self.useDistanceDelay:
            self.actual_position = self.detectorDistance.value * .001
            try:
                _motor = ExtraDataProxy(dc, self.motor.deviceId,
                                        data_selector_id=self.data_selector_id)
                key = self.motor.positionProperty.value
                unit_coef = self.motor.unitConversionCoef.value
                self.actual_position += unit_coef * _motor.get_value(key)
            except SourceNotFound:
                pass

    def _read_pp_decoder(self, dc):
        pp_src = dc[self.pp_decoder_id]
        trainId = pp_src['sase1/nPulses'].train_id_coordinates()
        bunch_pattern = np.zeros([trainId.size, 2700], dtype=np.uint32)
        n1 = pp_src['sase1/nPulses'].ndarray()
        p1 = pp_src['sase1/pulseIds'].ndarray()
        n2 = pp_src['sase2/nPulses'].ndarray()
        p2 = pp_src['sase2/pulseIds'].ndarray()
        n3 = pp_src['sase3/nPulses'].ndarray()
        p3 = pp_src['sase3/pulseIds'].ndarray()
        for i in range(trainId.size):
            bunch_pattern[i, p1[i, :n1[i]]] |= SA1.code
            bunch_pattern[i, p2[i, :n2[i]]] |= SA2.code
            bunch_pattern[i, p3[i, :n3[i]]] |= SA3.code
        return trainId, bunch_pattern

    def process(self, force=False):
        if not force and hasattr(self, 'r'):
            return self.r

        dc = self.dc
        self._read_run_values(dc)

        # bunch pattern
        if self.bunch_pattern_id in dc.instrument_sources:
            bp_keydata = dc[self.bunch_pattern_id, 'data.bunchPatternTable']
            bunch_pattern = bp_keydata.ndarray()
            trainId = bp_keydata.train_id_coordinates()
            bpsrc = "timeserver"
        elif self.pp_decoder_id in dc.control_sources:
            trainId, bunch_pattern = self._read_pp_decoder(dc)
            bpsrc = "pp_decoder"
        elif self._agipd.exp_type.value == 'XRay':
            raise(BunchPatternNotFound(
                f"Neither source '{self.bunch_pattern_id}' nor source "
                f"'{self.pp_decoder_id}' are found."))
        else:
            # only for dark runs
            bpsrc = "none"
            trainId = np.array(dc.train_ids, np.uint64)
            bunch_pattern = None

        # XGM
        if self.xgm_id not in dc.instrument_sources:
            xgm_flag = np.zeros(trainId.size, bool)
        else:
            xgm_en_keydata = dc[self.xgm_id, 'data.intensityTD']
            xgm_sig_keydata = dc[self.xgm_id, 'data.intensitySigmaTD']

            xgm_trainId = xgm_en_keydata.train_id_coordinates()
            xgm_sel = np.in1d(xgm_trainId, trainId)
            xgm_trainId = xgm_trainId[xgm_sel]

            xgm_intensityTD = xgm_en_keydata.ndarray()[xgm_sel]
            xgm_intensitySigmaTD = xgm_sig_keydata.ndarray()[xgm_sel]

            xgm_flag = np.in1d(trainId, xgm_trainId)
            xgm_map = np.roll(np.cumsum(xgm_flag), 1)
            xgm_map[0] = 0

        Output = namedtuple('Output', [
            'nFrame', 'detectorPulseId', 'nPulsePerFrame', 'energyPerFrame',
            'energySigma', 'dataFramePattern', 'masterPulseId', 'xgmPulseId'
        ])
        Result = namedtuple('Results', [
            'nPulse', 'nFrame', 'litFrames', 'nLitFrame', 'dataFrames',
            'nDataFrame', 'output', 'meta'
        ])
        Meta = namedtuple('Meta', ['litFrmDev', 'trainId'])
        n = trainId.shape[0]
        r = Result(
            nPulse=AVal(np.zeros(n, np.uint16)),
            nFrame=AVal(np.zeros(n, np.uint16)),
            nLitFrame=AVal(np.zeros(n, np.uint16)),
            litFrames=AVal(np.zeros([n, AGIPD_MAX_CELL], np.uint16)),
            nDataFrame=AVal(np.zeros(n, np.uint16)),
            dataFrames=AVal(np.zeros([n, AGIPD_MAX_CELL], np.uint16)),
            output=Output(
                nFrame=np.zeros(n, np.uint16),
                nPulsePerFrame=np.zeros([n, AGIPD_MAX_CELL], np.uint16),
                energyPerFrame=np.zeros([n, AGIPD_MAX_CELL], np.float32),
                energySigma=np.zeros([n, AGIPD_MAX_CELL], np.float32),
                dataFramePattern=np.zeros([n, AGIPD_MAX_CELL], bool),
                detectorPulseId=np.zeros([n, AGIPD_MAX_CELL], np.uint16),
                masterPulseId=np.zeros([n, 2700], np.uint16),
                xgmPulseId=np.zeros([n, 2700], np.uint16),
            ),
            meta=Meta(litFrmDev=f"offline/{bpsrc}", trainId=trainId),
        )

        Data = namedtuple('Data', ['data'])
        BunchPatternTable = namedtuple(
            'BunchPatternTable', ['bunchPatternTable'])
        XGM = namedtuple('XGM', ['intensityTD', 'intensitySigmaTD'])

        for no, tid in enumerate(trainId):
            if bunch_pattern is not None:
                bp = Data(Data(BunchPatternTable(AVal(bunch_pattern[no]))))

                if xgm_flag[no]:
                    xgm_no = xgm_map[no]
                    xgm = Data(Data(XGM(
                        intensityTD=AVal(xgm_intensityTD[xgm_no]),
                        intensitySigmaTD=AVal(xgm_intensitySigmaTD[xgm_no]),
                    )))
                else:
                    xgm = None

                (npulse, ncell,
                 agipd_pulse_ids, frames) = self.process_train(bp, xgm)
            else:
                (npulse, ncell,
                 agipd_pulse_ids, frames) = self.make_dark_run()

            # send pipeline (fast) data
            r.output.nFrame[no] = ncell
            r.output.detectorPulseId[no, :ncell] = agipd_pulse_ids
            r.output.nPulsePerFrame[no, :ncell] = frames.npulse
            r.output.energyPerFrame[no, :ncell] = frames.energy
            r.output.energySigma[no, :ncell] = frames.energy_std
            r.output.dataFramePattern[no, :ncell] = frames.hasdata
            nmaster = frames.master_ids.size
            r.output.masterPulseId[no, :nmaster] = frames.master_ids
            nxgm = frames.xgm_ids.size
            r.output.xgmPulseId[no, :nxgm] = frames.xgm_ids

            litframes = np.flatnonzero(frames.npulse)
            dataframes = np.flatnonzero(frames.hasdata)

            # update attributes (slow data)
            r.nPulse.value[no] = npulse
            r.nFrame.value[no] = ncell
            r.litFrames.value[no, :frames.nlit] = litframes
            r.nLitFrame.value[no] = frames.nlit
            r.dataFrames.value[no, :frames.ndata] = dataframes
            r.nDataFrame.value[no] = frames.ndata

        self.r = r
        return r

    def read(self):
        dc = self.dc
        inst_src = dc[self.dev_id + ':output']
        trainId = inst_src['data.nFrame'].train_id_coordinates()

        Output = namedtuple('Output', [
            'nFrame', 'detectorPulseId', 'nPulsePerFrame', 'energyPerFrame',
            'energySigma', 'dataFramePattern', 'masterPulseId', 'xgmPulseId'
        ])
        Result = namedtuple('Results', [
            'nPulse', 'nFrame', 'litFrames', 'nLitFrame', 'dataFrames',
            'nDataFrame', 'output', 'meta'
        ])
        Meta = namedtuple('Meta', ['litFrmDev', 'trainId'])
        out = Output(
            nFrame=inst_src['data.nFrame'].ndarray(),
            nPulsePerFrame=inst_src['data.nPulsePerFrame'].ndarray(),
            energyPerFrame=inst_src['data.energyPerFrame'].ndarray(),
            energySigma=inst_src['data.energySigma'].ndarray(),
            dataFramePattern=inst_src['data.dataFramePattern'].ndarray(),
            detectorPulseId=inst_src['data.detectorPulseId'].ndarray(),
            masterPulseId=inst_src['data.masterPulseId'].ndarray(),
            xgmPulseId=inst_src['data.xgmPulseId'].ndarray(),
        )

        n = trainId.size
        litframes = np.zeros([n, 352], dtype=np.uint16)
        dataframes = np.zeros([n, 352], dtype=np.uint16)
        for no, tid in enumerate(trainId):
            a = np.flatnonzero(out.nPulsePerFrame[no])
            litframes[no, :a.size] = a
            a = np.flatnonzero(out.dataFramePattern[no])
            dataframes[no, :a.size] = a

        r = Result(
            nPulse=AVal(out.nPulsePerFrame.sum(1)),
            nFrame=AVal(out.nFrame),
            nLitFrame=AVal((out.nPulsePerFrame != 0).sum(1)),
            litFrames=AVal(litframes),
            nDataFrame=AVal(out.dataFramePattern.sum(1)),
            dataFrames=AVal(dataframes),
            output=out,
            meta=Meta(litFrmDev=self.dev_id, trainId=trainId),
        )

        self.r = r
        return r

    def read_or_process(self):
        channel = self.dev_id + ':output'
        if channel in self.dc.instrument_sources:
            r = self.read()
            if len(r.meta.trainId):
                return r

        return self.process(force=True)


class AgipdLitFrameFinderConfigureMID:
    dev_id = "MID_EXP_AGIPD1M1/REDU/LITFRM"
    data_selector_id = 'MID_EXP_AGIPD1M1/MDL/DATA_SELECTOR'

    detector_control_id = "MID_EXP_AGIPD1M1/MDL/FPGA_COMP"
    bunch_pattern_id = "MID_RR_UTC/TSYS/TIMESERVER:outputBunchPattern"
    xgm_id = "SA2_XTD1_XGM/XGM/DOOCS:output"
    pp_decoder_id = "MID_RR_SYS/MDL/PULSE_PATTERN_DECODER"

    trigger = Trigger(
        deviceId="MID_EXP_SYS/TSYS/UTC-2-S4",
        name=AVal("backTrg3")
    )
    motor = Motor(
        deviceId="MID_AGIPD_MOTION/MDL/DOWNSAMPLER",
        positionProperty=AVal("t4EncoderPosition"),
        unitConversionCoef=AVal(1e-3),
    )

    instrumentName = AVal("MID")


class AgipdLitFrameFinderMID(AgipdLitFrameFinderConfigureMID,
                             AgipdLitFrameFinderOffline):
    def __init__(self, dc, dev_id=AgipdLitFrameFinderConfigureMID.dev_id,
                 ref_delays={}, align_method=None, **kwargs):
        super().__init__(self.dev_id, dc, ref_delays, align_method, **kwargs)


class AgipdLitFrameFinderConfigureSPB:
    dev_id = 'SPB_IRU_AGIPD1M1/REDU/LITFRM'
    data_selector_id = 'SPB_IRU_AGIPD1M1/MDL/DATA_SELECTOR'

    detector_control_id = 'SPB_IRU_AGIPD1M1/MDL/FPGA_COMP'
    bunch_pattern_id = 'SPB_RR_SYS/TSYS/X2TIMER2:outputBunchPattern'
    xgm_id = 'SPB_XTD9_XGM/XGM/DOOCS:output'
    pp_decoder_id = 'SPB_RR_SYS/MDL/BUNCH_PATTERN'

    trigger = Trigger(
        deviceId="SPB_EXP_SYS/TSYS/UTC_1_S3",
        name=AVal("backTrg3")
    )
    motor = Motor(
        deviceId="SPB_IRU_AGIPD1M/MOTOR/Z_STEPPER",
        positionProperty=AVal("encoderPosition"),
        unitConversionCoef=AVal(1e-3),
    )

    instrumentName = AVal('SPB')


class AgipdLitFrameFinderSPB(AgipdLitFrameFinderConfigureSPB,
                             AgipdLitFrameFinderOffline):
    def __init__(self, dc, dev_id=AgipdLitFrameFinderConfigureSPB.dev_id,
                 ref_delays={}, align_method=None, **kwargs):
        super().__init__(self.dev_id, dc, ref_delays, align_method, **kwargs)


class AgipdLitFrameFinderConfigureHED:
    dev_id = 'HED_EXP_AGIPD500K2G/REDU/LITFRM'

    detector_control_id = 'HED_EXP_AGIPD500K2G/MDL/FPGA_COMP'
    bunch_pattern_id = 'HED_RR_UTC/TSYS/TIMESERVER:outputBunchPattern'
    xgm_id = 'HED_XTD6_XGM/XGM/DOOCS:output'
    pp_decoder_id = 'SA2_BR_UTC/MDL/BUNCHPATTERN_DECODER'

    trigger = Trigger(
        deviceId="HED_EXP_SYS/TSYS/UTC-1-S2",
        name=AVal("backTrg3")
    )
    motor = Motor(
        deviceId="",
        positionProperty=AVal(""),
        unitConversionCoef=AVal(1),
    )

    instrumentName = AVal('HED')


class AgipdLitFrameFinderHED(AgipdLitFrameFinderConfigureHED,
                             AgipdLitFrameFinderOffline):
    def __init__(self, dc, dev_id=AgipdLitFrameFinderConfigureHED.dev_id,
                 ref_delays={}, align_method=None, **kwargs):
        super().__init__(self.dev_id, dc, ref_delays, align_method, **kwargs)


AgipdLitFrameFinderProfiles = {
    "MID": AgipdLitFrameFinderMID,
    "SPB": AgipdLitFrameFinderSPB,
    "HED": AgipdLitFrameFinderHED,
}


def find_litframe_finder_sources(dc):
    litfrm_finder_ids = [
        src for src, cls in (
            (src, dc.get_run_value(src, 'classId'))
            for src in dc.control_sources
        ) if cls == 'AgipdLitFrameFinder'
    ]
    return litfrm_finder_ids


def make_litframe_finder(inst, dc, dev_id=None, ref_delays={},
                         align_method=None):
    if inst not in AgipdLitFrameFinderProfiles:
        raise IndexError("Unexpected name of AgipdLitFrameFinder profile")

    litfrm_finder_cls = AgipdLitFrameFinderProfiles[inst]
    dev_ids = find_litframe_finder_sources(dc)

    if not dev_id:
        dev_id = litfrm_finder_cls.dev_id

    if dev_ids and dev_id not in dev_ids:
        dev_id = dev_ids[0]

    return litfrm_finder_cls(dc, dev_id, ref_delays=ref_delays,
                             align_method=align_method)



import numpy as np


def find_runs(x):
    """Find runs of consecutive items in an array."""
    n = len(x)
    if n == 0:
        return np.array([]), np.array([]), np.array([])

    loc_start = np.zeros(n, dtype=bool)
    loc_start[0] = True
    np.not_equal(x[:-1], x[1:], out=loc_start[1:])
    starts = np.flatnonzero(loc_start)

    values = x[starts]
    lengths = np.diff(np.append(starts, n))

    return values, starts, lengths


def find_ranges(ix):
    """Find sequences in an array on integers,
       which can be defined with limits and step
    """
    if len(ix) == 0:
        return []
    rng_list = []
    first, last, step, count = 0, 0, 1, 0
    for t in ix:
        if count == 0:
            first = t
            count = 1
        elif count == 1:
            count = 2
            step = t - last
        else:
            s = t - last
            if step == s:
                count += 1
            elif count == 2:
                rng_list.append((first, first, 1, 1))
                first = last
                step = s
                count = 2
            else:
                rng_list.append((first, last, step, count))
                first = t
                count = 1
        last = t

    rng_list.append((first, last, step, count))
    return sorted(rng_list)


def find_ranges_of_patterns(nptrn, ix, tid):
    """Find train ranges for a number of patterns"""
    patterns = {}
    for a in range(nptrn):
        patterns[a] = find_ranges(tid[np.flatnonzero(ix == a)])

    return patterns


def compress_ranges_of_patterns(patterns, tid):
    """Merge the subsequent ranges if they were splitted
       from bigger range due to a few missed trains.
    """
    patterns_compressed = []
    for a, ranges in sorted(patterns.items(), key=lambda v: v[1][0]):
        rng_list = []

        for rng in ranges:
            rng2 = tuple(list(rng) + [0])
            s2 = rng2[2]

            while rng_list:
                rng1 = rng_list[-1]
                s1 = rng1[2]

                if (s1 == 0 and s2 == 0) or (s1 != s2 and s1 * s2 != 0):
                    break

                delta = int(rng2[0]) - int(rng1[1])
                s = max(s1, s2)

                nmissed = -1
                if (delta % s) == 0:
                    missed_trains = np.arange(rng1[1] + s, rng2[0], s,
                                              dtype=int)
                    if np.all(np.in1d(missed_trains, tid, invert=True)):
                        nmissed = len(missed_trains)

                if nmissed != -1:
                    rng2 = (rng1[0], rng2[1], s, rng1[3] + rng2[3],
                            rng1[4] + rng2[4] + nmissed)
                    rng_list.pop()
                else:
                    break

            rng_list.append(rng2)

        patterns_compressed.append((a, rng_list))

    return patterns_compressed


def expand_ranges_of_cells(cell_ranges):
    """Expand ranges to the sequence of single numbers
       if they define short sequence
    """
    int_expanded = []
    for f, l, s, n in cell_ranges:
        if n == 1:
            int_expanded.append((f,))
        elif n == 2:
            int_expanded += [(f,), (l,)]
        elif s <= 1:
            int_expanded.append((f, l + 1))
        elif n == 3:
            int_expanded += [(f,), (f + s,), (l,)]
        else:
            int_expanded.append((f, l + 1, s))

    return int_expanded



import os
import sys
import csv
import argparse
import numpy as np

from extra_data import RunDirectory
from extra_data.exceptions import SourceNameError
from extra_data.read_machinery import find_proposal
from extra_redu import (
    make_litframe_finder, BunchPatternNotFound, DetectorNotFound,
    ReferenceDelay, FrameSelection,
)
from extra_redu.base import EVENT_CODE
from extra_redu.base import DESTINATION


def get_npulse_xgm(dc, xgm_id, dst):
    dev_id = xgm_id.partition(':')[0]
    if dev_id not in dc.control_sources:
        raise SourceNameError(dev_id)
    keydata = dc[dev_id, f"pulseEnergy.numberOf{dst}BunchesActual"]
    location = dc.get_run_value(dev_id, 'location')
    return keydata.train_id_coordinates(), keydata.ndarray(), location


def get_npulse_pp(dc, pp_id, dst):
    if pp_id not in dc.control_sources:
        raise SourceNameError(pp_id)
    keydata = dc[pp_id, f"{dst}.nPulses"]
    return keydata.train_id_coordinates(), keydata.ndarray()


def get_npulse_bp(dc, ts_id, dst):
    dev_id, _, channel = ts_id.partition(':')
    if not channel:
        channel = "outputBunchPattern"
    src = dev_id + ':' + channel
    if src not in dc.instrument_sources:
        raise SourceNameError(dev_id)
    bp = dc[src, 'data.bunchPatternTable'].ndarray()
    tid = dc[src, 'data.trainId'].ndarray()
    return tid, np.sum(dst.get_instrument_pulse_mask(bp), 1)


def get_most_frequently(trains, tid, npulse):
    ix = np.flatnonzero(np.in1d(tid, trains))
    npulse_uniq, count = np.unique(npulse[ix], return_counts=True)
    return npulse_uniq[np.argmax(count)]


def parse_slice(optval):
    """Parses modules cmdline option"""
    rng = set()
    for t in optval.split(','):
        a, _, b = t.partition('-')
        if b == '':
            rng.add(int(a))
        else:
            rng.update(range(int(a), int(b) + 1))

    return sorted(rng)


def parse_delay(optstr):
    event_name, _, params = optstr.partition(',')
    event = EVENT_CODE[event_name.strip()]
    types = [int, int, int, float, float]
    args = [type_cast(tk) for tk, type_cast in zip(params.split(','), types)]
    return (event, ReferenceDelay(*args))


def main():
    parser = argparse.ArgumentParser(
        description='Generate litframe report for a proposal.')
    parser.add_argument('proposal', type=int, help='proposal Id')
    parser.add_argument('-o', '--csv', type=str,
                        help='the name of output CSV-file')
    parser.add_argument('-r', '--runs', type=parse_slice,
                        help='the slice of runs, e.g. 1-3,5,7-10')
    parser.add_argument('-d', '--delay', type=parse_delay, action='append',
                        help="the reference delay as: static|dinamic,<delay>,"
                        "<pulse>,<cell>,<rep.rate>,<distance>")
    parser.add_argument('-a', '--align-method', type=str,
                        choices=['by reference delay', 'by first pulse'],
                        help="force alignment method")
    parser.add_argument('-p', '--process', action='store_true',
                        default=False, help="force reprocessing")

    args = parser.parse_args()

    propno = args.proposal
    csvname = f"litfrm_p{propno:06}.csv" if args.csv is None else args.csv
    ref_delays = dict(args.delay) if args.delay else {}
    align_method = args.align_method
    reprocess = args.process or (align_method is not None)

    propdir = find_proposal(f"p{propno:06d}")
    inst = propdir.split('/')[4]
    dst = DESTINATION[inst]
    print(f"Proposal No: {propno}")
    print(f"Proposal directory: {propdir}")
    print(f"Instrument: {inst}")
    print(f"Align method: {align_method}")
    if args.delay:
        print("Default reference delays:")
        for event, ref in ref_delays.items():
            print(f"  {event:3d}:{ref}")
    print()

    runs = sorted([
        int(fn[1:]) for fn in os.listdir(os.path.join(propdir, "raw"))
        if fn[0] == 'r' and fn[1:].isdigit()
    ])
    if args.runs is not None:
        runs = [runno for runno in runs if runno in args.runs]

    f = open(csvname, 'w', newline='')
    fieldnames = ['runno', 'source', 'runtype', 'pattern_no', 'range_serial',
                  'train_range', 'nmissed_trains', 'npulse_bp', 'npulse_pp',
                  'npulse_xgm', 'npulse_exposed', 'ndataframe', 'nframe_total',
                  'litframe_slice', 'super_pattern_slice', 'xgm_location',
                  'exception']
    fieldlabels = ['Run No', 'Source', 'AGIPD Run Type', 'Pattern No',
                   'Train Range Serial', 'Train Range',
                   'Number of missed trains', 'Number of pulses at Timeserver',
                   'Number of pulses at PPD', 'Number of pulses at XGM',
                   'Number of Exposed Pulses', 'Number of Frames with Data',
                   'Total Number of Frames', 'Frame Slice',
                   'Common Frame Slice', 'XGM Location', 'Exception']
    writer = csv.DictWriter(
        f, fieldnames=fieldnames, delimiter=';', quotechar='"',
        quoting=csv.QUOTE_NONNUMERIC
    )
    writer.writerow(dict(zip(fieldnames, fieldlabels)))

    for runno in runs:
        rundir = os.path.join(propdir, "raw", f"r{runno:04}")
        dc = RunDirectory(rundir)

        try:
            litfrm = make_litframe_finder(inst, dc, ref_delays=ref_delays,
                                          align_method=align_method)
            if reprocess:
                r = litfrm.process()
            else:
                r = litfrm.read_or_process()

            runtype = litfrm._detector.patternType.value

            try:
                tid_xgm, npulse_xgm, xgm_loc = get_npulse_xgm(
                    dc, litfrm.xgm_id, dst.xgm_key)
            except SourceNameError:
                tid_xgm, npulse_xgm = None, None
            try:
                tid_pp, npulse_pp = get_npulse_pp(
                    dc, litfrm.pp_decoder_id, dst.pp_key)
            except SourceNameError:
                tid_pp, npulse_pp = None, None
            try:
                tid_bp, npulse_bp = get_npulse_bp(
                    dc, litfrm.bunch_pattern_id, dst)
            except SourceNameError:
                tid_bp, npulse_bp = None, None

            sel = FrameSelection(r)
            for r in sel.report():
                frmintf = ', '.join([
                    ':'.join([str(n) for n in slc])
                    for slc in r['litframe_slice']
                ])
                supperptrn_intf = ', '.join([
                    ':'.join([str(n) for n in slc])
                    for slc in r['super_pattern_slice']
                ])
                trsintf = ':'.join([str(n) for n in r['train_range']])
                recfrm = r.copy()
                tid = range(*r['train_range'])
                if tid_xgm is not None:
                    recfrm['npulse_xgm'] = get_most_frequently(
                        tid, tid_xgm, npulse_xgm)
                    recfrm['xgm_location'] = xgm_loc
                if tid_pp is not None:
                    recfrm['npulse_pp'] = get_most_frequently(
                        tid, tid_pp, npulse_pp)
                if tid_bp is not None:
                    recfrm['npulse_bp'] = get_most_frequently(
                        tid, tid_bp, npulse_bp)
                recfrm['litframe_slice'] = frmintf
                recfrm['super_pattern_slice'] = supperptrn_intf
                recfrm['runno'] = runno
                recfrm['runtype'] = runtype
                writer.writerow(recfrm)
                if r['range_serial'] == 0:
                    print("{runno:4d} {source:30s} {runtype:6s}".format(
                        **recfrm), end='')
                else:
                    print(" "*42, end='')
                print(
                    ("{pattern_no:2d} {trsintf:25s} {npulse_exposed:4d} "
                     "{ndataframe:3d} {nframe_total:3d} [{litframe_slice}]"
                     ).format(trsintf=trsintf, **recfrm))
        except BunchPatternNotFound:
            runtype = litfrm._detector.patternType.value
            errmsg = f"BunchPattern not found ({runtype})"
            writer.writerow(dict(runno=runno, exception=errmsg))
            print(f"{runno:4d} - skip: {errmsg}")
        except DetectorNotFound:
            errmsg = "Detector not found"
            writer.writerow(dict(runno=runno, exception=errmsg))
            print(f"{runno:4d} - skip: {errmsg}")

    f.close()


if __name__ == "__main__":
    sys.exit(main())



import numpy as np
from enum import IntEnum
from .utils import (
    find_runs, find_ranges_of_patterns, compress_ranges_of_patterns,
    expand_ranges_of_cells, find_ranges,
)


def row_selection(ptrn, rowsize=32):
    """Expand cell selection to row selection"""
    single = ptrn.ndim == 1
    nfrm = ptrn.size if single else ptrn.shape[1]
    nrow = nfrm // rowsize + bool(nfrm % rowsize)
    padded_size = nrow * rowsize
    npad = padded_size - nfrm
    ptrn_cm = np.pad(ptrn, (0, npad)).reshape(-1, nrow, rowsize)
    ptrn_cm[:, :] = np.any(ptrn_cm, axis=2, keepdims=True)
    ptrn_cm = ptrn_cm.reshape(-1, padded_size)[:, :nfrm]
    return ptrn_cm.ravel() if single else ptrn_cm


class SelType(IntEnum):
    CELL = 0
    ROW = 1
    SUPER_CELL = 2
    SUPER_ROW = 3


class FrameSelection:

    def __init__(self, r, guess_missed=True, crange=None,
                 energy_threshold=-1000, select_litframes=False):
        self.tid = r.meta.trainId
        if energy_threshold != -1000:
            self.energy = r.output.energyPerFrame
            self.missed_xgm_intensity = 10000

        self.dev = r.meta.litFrmDev
        self._find_pattern_ranges(r, guess_missed, select_litframes)
        self._make_selection(guess_missed, crange, energy_threshold)

    def _find_pattern_ranges(self, r, guess_missed=True,
                             select_litframes=False):
        """Find unique patterns and their train ranges"""
        self.ranges = []
        self.datafrm = []
        self.npls = []
        self.npls_ix = []
        self.run_slice = []

        # find run of trains with fixed number of frames
        nfrm_uniq, run_starts, ntrn_in_run = find_runs(r.output.nFrame)

        self.nfrmrun = len(nfrm_uniq)
        self.nfrm_uniq = nfrm_uniq
        self.tid_type = r.meta.trainId.dtype

        self.ntrain = np.zeros(self.nfrmrun, dtype=int)
        for i in range(self.nfrmrun):
            nfrm = nfrm_uniq[i]
            i0 = run_starts[i]
            iN = i0 + ntrn_in_run[i]

            if select_litframes:
                datafrm = r.output.nPulsePerFrame[i0:iN, :nfrm] > 0
            else:
                datafrm = r.output.dataFramePattern[i0:iN, :nfrm]
            tid = r.meta.trainId[i0:iN]

            # find unique patterns
            npls_uniq, npls_first, npls_ix = np.unique(
                r.output.nPulsePerFrame[i0:iN, :nfrm],
                return_index=True, return_inverse=True, axis=0
            )
            # make ranges of the same patterns
            ranges = find_ranges_of_patterns(npls_uniq.shape[0], npls_ix, tid)
            ranges = compress_ranges_of_patterns(ranges, tid)

            # compress data frame patterns
            ptrn = datafrm[npls_first, :]

            # accumulate normalised number of trains in the frame run
            if guess_missed:
                ntrn_nrm = 0
                for _, range_list in ranges:
                    ntrn_nrm += sum(ntrn + missed
                                    for _, _, _, ntrn, missed in range_list)
                self.ntrain[i] = ntrn_nrm
            else:
                self.ntrain[i] = ntrn_in_run

            self.ranges.append(ranges)
            self.datafrm.append(ptrn)
            self.npls.append(npls_uniq)
            self.npls_ix.append(npls_ix)
            self.run_slice.append(slice(i0, iN))

    def _make_selection(self, guess_missed=True,
                        crange=None, energy_threshold=-1000):
        """Make frame selection from lit frame patterns"""
        ntrain = np.sum(self.ntrain)
        nframe = np.sum(self.ntrain * self.nfrm_uniq)
        self.trains = np.zeros(ntrain, dtype=self.tid_type)
        self.cell_flags = np.zeros(nframe, dtype=bool)
        self.row_flags = np.zeros(nframe, dtype=bool)
        self.cell_super = np.zeros(nframe, dtype=bool)
        self.row_super = np.zeros(nframe, dtype=bool)
        self.empty_trains = np.zeros(ntrain, dtype=bool)
        self.first = np.zeros(ntrain, dtype=int)
        self.count = np.zeros(ntrain, dtype=int)
        self.flags = [self.cell_flags, self.row_flags,
                      self.cell_super, self.row_super]
        one = self.tid_type.type(1)

        i0 = 0
        k0 = 0
        for i in range(self.nfrmrun):
            ntrn_in_run = self.ntrain[i]
            nfrm = self.nfrm_uniq[i]
            s = self.run_slice[i]

            if crange is None:
                cell_mask = np.ones(nfrm, bool)
            else:
                cell_mask = np.zeros(nfrm, bool)
                cell_mask[crange] = True

            ptrn = self.datafrm[i] & cell_mask[None, :]

            # make row selection for common mode correction
            ptrn_cm = row_selection(ptrn)

            # find empty pattern
            ptrn_empty = np.sum(ptrn, axis=1) == 0

            # find the super pattern
            superptrn = np.any(ptrn, axis=0)
            superptrn_cm = row_selection(superptrn)

            iN = i0 + ntrn_in_run * nfrm
            kN = k0 + ntrn_in_run
            if guess_missed:
                # make normalised trains
                ranges = self.ranges[i]

                tid_nrm = np.zeros(ntrn_in_run, dtype=self.tid_type)
                ptrn_ix = np.zeros(ntrn_in_run, dtype=int)
                j0 = 0
                for ptrn_id, range_list in ranges:
                    for t0, tN, step, ntrn, missed in range_list:
                        # make train list
                        jN = j0 + ntrn + missed
                        tid_nrm[j0:jN] = np.arange(
                            t0, tN + one, step, dtype=self.tid_type)
                        ptrn_ix[j0:jN] = ptrn_id
                        j0 = jN

                sorted_ix = np.argsort(tid_nrm)
                self.trains[k0:kN] = tid_nrm[sorted_ix]

                ix = ptrn_ix[sorted_ix]
            else:
                # use original trains
                self.trains[k0:kN] = self.tid[s]
                ix = self.npls_ix[i]

            # expand and stack patterns in pulse resolved format
            self.cell_flags[i0:iN] = ptrn[ix, :].ravel()
            self.row_flags[i0:iN] = ptrn_cm[ix, :].ravel()
            self.empty_trains[k0:kN] = ptrn_empty[ix]

            # apply energy threshold
            if energy_threshold != -1000:
                tid = self.tid[s]
                en = self.energy[s, :nfrm]
                xgm_flag = np.ones([ntrn_in_run, nfrm], bool)
                xgm_ix = np.flatnonzero(np.in1d(tid, self.trains[k0:kN]))
                xgm_flag[xgm_ix, :] = ((en > energy_threshold) |
                                       (en == self.missed_xgm_intensity))

                self.cell_flags[i0:iN] = (
                    self.cell_flags[i0:iN] & xgm_flag.ravel())

            self.cell_super[i0:iN] = np.tile(superptrn, ntrn_in_run)
            self.row_super[i0:iN] = np.tile(superptrn_cm, ntrn_in_run)

            self.first[k0:kN] = np.arange(i0, iN, nfrm, dtype=int)
            self.count[k0:kN] = np.full(ntrn_in_run, nfrm, dtype=int)

            i0 = iN
            k0 = kN

    def litframes_on_trains(self, train_sel, nfrm, frame_idx, selection_types):
        """Return lit-frame selection for requested trains"""
        ntrain = len(train_sel)
        _, ix_avl, ix_sel = np.intersect1d(
            self.trains, train_sel, assume_unique=True, return_indices=True)

        train_idx = np.full(ntrain, -1, dtype=int)
        train_idx[ix_sel] = ix_avl

        nsel = len(selection_types)
        nfrm_total = np.sum(nfrm)
        flags = [np.zeros(nfrm_total, dtype=bool) for _ in range(nsel)]
        count = [np.zeros(ntrain, dtype=int) for _ in range(nsel)]

        j0 = 0
        for trn_no, trn_ix in enumerate(train_idx):
            jN = j0 + nfrm[trn_no]
            if trn_ix == -1 or self.empty_trains[trn_ix]:
                for i in range(nsel):
                    count[i][trn_no] = 0
            else:
                fix = frame_idx[j0:jN]
                # what to do if frame index is out of range?
                # clean or raise exception
                i0 = self.first[trn_ix]
                iN = i0 + self.count[trn_ix]
                for i, sel_type in enumerate(selection_types):
                    flags_trn = self.flags[sel_type][i0:iN][fix]
                    flags[i][j0:jN] = flags_trn
                    count[i][trn_no] = np.sum(flags_trn)
            j0 = jN

        return zip(flags, count)

    def filter_trains(self, train_sel, drop_empty=True):
        """Filters out trains that will not be processed"""
        good_trains = (self.trains[~self.empty_trains]
                       if drop_empty else self.trains)
        return train_sel[np.in1d(train_sel, good_trains)]

    def report(self):
        """Generate litframe report"""
        one = self.tid_type.type(1)

        rep = []
        ptrn_no = 0
        serial = 0
        for i in range(self.nfrmrun):
            nfrm = self.nfrm_uniq[i]
            ranges = self.ranges[i]
            datafrm = self.datafrm[i]
            npls_uniq = self.npls[i]

            # find the super pattern
            superptrn = np.any(datafrm, axis=0)
            superptrn_int = expand_ranges_of_cells(
                find_ranges(np.flatnonzero(superptrn)))

            for ptrn_id, range_list in ranges:
                for t0, tN, step, count, missed in range_list:
                    trains = ((t0, tN + one, step) if step > 1
                              else (t0, tN + one))

                    ptrn = datafrm[ptrn_id, :]
                    ndata = ptrn.sum()
                    npls = npls_uniq[ptrn_id, :].sum()

                    frmint = expand_ranges_of_cells(
                        find_ranges(np.flatnonzero(ptrn)))
                    rep.append(dict(
                        source=self.dev, pattern_no=ptrn_no,
                        range_serial=serial, train_range=trains,
                        nmissed_trains=missed, npulse_exposed=npls,
                        ndataframe=ndata, nframe_total=nfrm,
                        litframe_slice=frmint,
                        super_pattern_slice=superptrn_int,
                    ))
                    serial += 1

                ptrn_no += 1

        return rep



cdef extern from "rounding.h":
    float errb_round_f(float arg, int mnbits, int abspw)
    void errb_round_vec(int n, const float *v, int incv, float *r, int incr, int mnbits, int abspw)
    void errb_round_arr(int ndim, int *n, const float *v, const int *incv, 
                        float *r, const int *incr, int mnbits, int abspw)



from .agipd_offline import (
    AgipdLitFrameFinderOffline, AgipdLitFrameFinderMID,
    AgipdLitFrameFinderSPB, AgipdLitFrameFinderHED,
    LitFrameFinderError, BunchPatternNotFound, DetectorNotFound,
    TriggerNotFound, NoReferenceDelay, make_litframe_finder
)
from .selection import FrameSelection, SelType



from euxfel_bunch_pattern import (
    DESTINATION_MASK, PHOTON_LINE_DEFLECTION, DESTINATION_T4D, DESTINATION_T5D
)


class ParamDecoder:
    """Decodes Bunch Pattern"""

    @classmethod
    def get_instrument_pulse_mask(cls, bunch):
        return (bunch & cls.mask) == cls.code

    @classmethod
    def get_xgm_pulse_mask(cls, bunch):
        xgm_pulses = (bunch & cls.mask_xgm) == cls.code_xgm
        inst_pulses = (bunch & cls.mask) == cls.code
        return inst_pulses[xgm_pulses]

    @classmethod
    def get_xgm_pulse_mask_in_file(cls, bunch, xgm_maxlen=1000):
        xgm_pulses = (bunch & cls.mask_xgm) == cls.code_xgm
        inst_pulses = (bunch & cls.mask) == cls.code
        return inst_pulses[xgm_pulses]


class SA1(ParamDecoder):
    """SASE 1 Bunch Pattern Decoder"""
    mask = PHOTON_LINE_DEFLECTION | DESTINATION_MASK
    code = DESTINATION_T4D
    mask_xgm = DESTINATION_MASK
    code_xgm = DESTINATION_T4D
    event_code = {"dynamic": 180, "static": 16}
    xgm_key = "Sa1"
    pp_key = "sase1"


class SA2(ParamDecoder):
    """SASE 2 Bunch Pattern Decoder"""
    mask = DESTINATION_MASK
    code = DESTINATION_T5D
    mask_xgm = DESTINATION_MASK
    code_xgm = DESTINATION_T5D
    event_code = {"dynamic": 181, "static": 16}
    xgm_key = ""
    pp_key = "sase2"


class SA3(ParamDecoder):
    """SASE 3 Bunch Pattern Decoder"""
    mask = PHOTON_LINE_DEFLECTION | DESTINATION_MASK
    code = PHOTON_LINE_DEFLECTION | DESTINATION_T4D
    mask_xgm = DESTINATION_MASK
    code_xgm = DESTINATION_T4D
    event_code = {"dynamic": 182, "static": 16}
    xgm_key = "Sa3"
    pp_key = "sase3"


DESTINATION = {
    'SPB': SA1, 'FXE': SA1,
    'MID': SA2, 'HED': SA2,
    'SQS': SA3, 'SCS': SA3,
}



import cython

import numpy as np
cimport numpy as np

from . cimport crounding

ctypedef np.float32_t float_t


@cython.boundscheck(False)
@cython.wraparound(False)
def errb_round(object inp, int mnbits=23, int abspw=-150, object out=None):
    """
    Round values in array up to absolute and relative precision.

    This function rounds elements of arrays up to relative and absolute precision.
    Relative precision defined as the number of significant mantissa bits given by
    the `mnbits` parameter. Absolute precision defined as the difference between
    two consecutive representative numbers and given as the power of two by the
    parameter `abspw`.

    :param inp:     An input array of single precision numbers
    :param mnbits:  The number of significant mantissa bits
    :param abspw:   The absolute precision given as the power of two,
                    actual precision is 2**abspw, e.g. 2**(-1) = 0.5
    :param out:     An output array with the same shape as the input array

    :return out:    An output array with rounded numbers
    """
    if inp.dtype != np.float32:
        raise ValueError(f"Input buffer invalid dtype, expected 'float32' but got '{inp.dtype}'")

    cdef int ndim = inp.ndim

    if out is None:
        out = np.zeros(inp.shape, np.float32)
    else:
        if out.dtype != np.float32:
            raise ValueError(f"Output buffer invalid dtype, expected 'float32' but got '{out.dtype}'")
        
    cdef np.ndarray inpbuf = inp
    cdef np.ndarray outbuf = out
    
    cdef float* a = <float *> inpbuf.data
    cdef float* r = <float *> outbuf.data
    
    cdef int n[32], inca[32], incr[32]
    cdef int dim
    
    cdef bint shape_mismatch = False
    for dim in range(ndim):
        n[dim] = inpbuf.shape[dim]
        shape_mismatch = shape_mismatch or (inpbuf.shape[dim] != outbuf.shape[dim])
        inca[dim] = inpbuf.strides[dim] / inpbuf.itemsize
        incr[dim] = outbuf.strides[dim] / outbuf.itemsize

    if shape_mismatch:
        raise ValueError("Input and output arrays have different shape")
        
    crounding.errb_round_arr(ndim, n, a, inca, r, incr, mnbits, abspw)
    return out


def errb_round_flat(np.ndarray inp, int mnbits=23, int abspw=-150, np.ndarray out=None):
    """
    Round values in array up to absolute and relative precision.

    This function rounds elements of arrays up to relative and absolute precision.
    Relative precision defined as the number of significant mantissa bits given by
    the `mnbits` parameter. Absolute precision defined as the difference between
    two consecutive representative numbers and given as the power of two by the
    parameter `abspw`.
    
    This version flattens arrays and can copy the buffer therefore.
    Do not specify sliced views to avoid it.

    :param inp:     An input array of single precision numbers
    :param mnbits:  The number of significant mantissa bits
    :param abspw:   The absolute precision given as the power of two,
                    actual precision is 2**abspw, e.g. 2**(-1) = 0.5
    :param out:     An output array with the same shape as the input array

    :return out:    An output array with rounded numbers
    """
    if inp.dtype != np.float32:
        raise ValueError(f"Buffer dtype mismatch, expected 'float32' but got '{inp.dtype}'")

    shape = (<object> inp).shape
    inp_view = inp.ravel()

    if out is None:
        if inp_view.flags['OWNDATA']:
            # `inp_view` is a copy, we may use it as output buffer
            out = inp_view.reshape(shape)
            out_view = inp_view
        else:
            # else allocate a new buffer
            out = np.zeros_like(inp)
            out_view = out.ravel()
    else:
        if inp.size != out.size:
            raise ValueError("Input and output arrays have different size")

        out_view = out.ravel()
        # Make sure we are still writing to the user's buffer
        if out_view.flags['OWNDATA']:
            raise ValueError("Output buffer cannot be flattened without copying")

    cdef float[:] a = inp_view
    cdef float[:] r = out_view

    cdef int inca = a.strides[0] // a.itemsize
    cdef int incr = r.strides[0] // r.itemsize

    crounding.errb_round_vec(a.size, &a[0], inca, &r[0], incr, mnbits, abspw)
    return out
